//                              -*- Mode: C++ -*-
//: EMClusterTree.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 3/4/97
// Last Update ........
//

#include "care/EMClusterTree.hh"
#include "care/EAClusterAnalyzer.hh"
#include "care/EDClusterModifier.hh"
#include "arte/ArteMessage.hh"
#include "care/ELLUTsManager.hh"
#include "care/ELCellGrid.hh"
#include <algorithm>



const unsigned int MAXCELLS = 6000;
const unsigned int MAXTRACKS =500;
const unsigned int MAXFRACT = 1000;
//////////////  Class EMClusterTree definitions //////////////////////////////

// Default constructor
EMClusterTree::EMClusterTree() : EMCluster(), 
  subCluster(MAXCELLS), clusCell(MAXCELLS), trackOnEcal(MAXTRACKS),
  cellFrac(MAXFRACT){
  reset();
}


//-- Destructor
//   A virtual destructor is needed
EMClusterTree::~EMClusterTree(){
  // start destroying each object in the list
  // then remove the list
}




//--
// Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
// that are object with definite analyzing properties. 
// EMClusterMembers MUST ignore these properties: in this way a loose coupling
// between EMClusterMembers and EAClusterAnalyzers is achieved and both
// can vary independently during program evolution.
// Examples of EAClusterAnalyzers are those that should return
// total energy of a cluster, the Center of Gravity, the Number of Cells.
void EMClusterTree::analyzeMe(EAClusterAnalyzer & analyzer) const {
  //
  analyzer.doAnalysis(this);
}

void EMClusterTree::modifyMe(EDClusterModifier & modifier){
  //
  modifier.doModify(this);
}



//--
//  Dump for debugging
void EMClusterTree::dump() const {
  arteMsg<<"****************************************************";
  arteMsg.allp();
  arteMsg<<"Cluster List Hypotheses: (H:"<< getHypotheses()
      <<", T:"<<getTracks() 
      <<", C:"<<getChargHyp()
      <<", N:"<<getNeuHyp()<<") ";
  arteMsg.allp();
  EMCluster::dump();
  arteMsg<<"****************************************************";
  arteMsg.allp();
  arteMsg<<"  Dumping All Known Tracks On ECAL :"<<howManyTracks();
  arteMsg.allp();
  for(std::vector<const EMTrackOnEcal>::const_iterator ti = beginTrackList();
      ti != endTrackList(); ++ti){
    ti->dump();  
  }
  arteMsg<<"****************************************************";
  arteMsg.allp();
}


//-- Reset all info in the cluster
void  EMClusterTree::reset(){
  EMCluster::reset();
  nextClusIdx =0;
  nextCellIdx = 0;
  nextTrackIdx = 0;
  nextFracIdx = 0;
  // Initialize LUT for cluster pointers
  // assume there are 6000 cells
  if(whichHitCellArray.capacity()<MAXCELLS){
    whichHitCellArray.reserve(MAXCELLS);
  }
}

//-- Start preparing to receive "size" cells
void  EMClusterTree::start(unsigned int size){
  unsigned int ncells = MAXCELLS;
  reset();
  std::fill_n(whichHitCellArray.begin(), ncells, (EMHitCell *)0);
  //
  // Initialize SubCluster and ClusCell arrays
  std::fill(subCluster.begin(),subCluster.end(),EMCluster());
  std::fill(clusCell.begin(),clusCell.end(),EMHitCell());
  std::fill(cellFrac.begin(),cellFrac.end(),EMCellFraction());

}


//--
// start a new cluster with a given cell.
// the passed cell is actually MOVED
EMCluster * EMClusterTree::newCluster(EMClusterMember * clusmem){
  // std::cout<<"start..a new cluster "<<nextClusIdx ;
  subCluster[nextClusIdx].start(clusmem);
  // std::cout<<"  after start"<<std::endl;
  EMCluster * newClus = &subCluster[nextClusIdx];
  nextClusIdx++;
  return  newClus;
}


//--
/// get a pointer to the next cell to clusterize
EMHitCell*  EMClusterTree::newHitCell(CellIdx_t currIdx){
  EMHitCell* hitcell = &clusCell[nextCellIdx];
  whichHitCellArray[currIdx.index()] = &clusCell[nextCellIdx];
// update whichHitCellArray
  nextCellIdx++;
  return hitcell;
}

//--
// Store parameters of a track  (called somewhere before caresecondstage) 
void EMClusterTree::newTrack(TrackInfo & trackInfo){
  // fill the parameters for the current track
 
  if(nextTrackIdx<trackOnEcal.size()){
    trackOnEcal[nextTrackIdx++].fill(trackInfo);
  } else if( nextTrackIdx<MAXTRACKS ){
    trackOnEcal.push_back(EMTrackOnEcal());
    trackOnEcal[nextTrackIdx++].fill(trackInfo);
  } else {
    static int StoreWarn = 0x0ECA0006;
    static int neve = 0;

    if( neve!=ArteTable<EVHD>::begin()->evt ){
      arteMsg<<" Store Track: more than "<<MAXTRACKS
 	     <<" tracks.";
      arteMsg.warn(StoreWarn,100);
      neve = ArteTable<EVHD>::begin()->evt;
    } 
  }
}


//--
/// get the space for a new cell fraction 
EMCellFraction* EMClusterTree::newCellFraction(EMHitCell* cell, 
					       Energy_t fraction){
  EMCellFraction* where = 0;
  // fill the parameters for the current cell fraction
  if(nextFracIdx<cellFrac.size()){
    where = cellFrac[nextFracIdx++].pointToCell(cell, fraction);
  } else if( nextFracIdx<MAXFRACT ){
    cellFrac.push_back(EMCellFraction(cell, fraction));
    where = &cellFrac.back();
  } else {
    static int FracWarn = 0x0ECA0007;
    static int nevef = 0;
    if( nevef!=ArteTable<EVHD>::begin()->evt ){
      arteMsg<<" Store Cell Fraction: more than "<<MAXFRACT
             <<" fractional cells.";
      arteMsg.warn(FracWarn,100);
      nevef = ArteTable<EVHD>::begin()->evt;    
    } 
  }
  return where;
}


//--
// return the base cluster in which the point on ECAL is
EMHitCell* EMClusterTree::whichHitCell(Coord_t x, Coord_t y) const{

  ELLUTsManager LM;
  ELCellGrid::ErrorTypes ierr;
  // get the number of nearest neibouring cells of the current one
  CellIdx_t ncell = LM.cellsGrid().cellIdx(x,y,ierr);  
  EMHitCell* where = 0;
  if( ierr== ELCellGrid::OK){
    where = whichHitCellArray[ncell.index()];
  }
  return where;
}
