//                              -*- Mode: C++ -*-
//: EMTrackOnEcal.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 9/5/97
// Last Update ........
//

#ifndef EMTRACKONECAL_HH_
#define EMTRACKONECAL_HH_


#include "care/EMClusterMember.hh"
#include "care/EMHitCell.hh"
#include "care/ECLastHit.hh"

class EMClusterTree;


/**
class  EMTrackOnEcal.
  Concrete derived class for the hierarchical cluster tree used in
  pattern recognition of ECAL. This class is a "leaf": it does not have
  children, it cannot receive an Add message.
  It is used to store the relation between an extrapolated track on the
  calorimeter and the EMHitCell in which the extrapolated point is. 
  
*/


class EMTrackOnEcal : public EMClusterMember {
public:
  /**@name Constructor and Destructor */
  //@{
  /** Constructor:
   The default constructor should be used only by the HCBuilder class. */
  EMTrackOnEcal();
  /// Destructor: a virtual destructor is needed.
  virtual ~EMTrackOnEcal();
  //@}


  /**@name Status access functions */
  //@{
  /// Return a flag to tell if this EMClusterMember IS a leaf.
  virtual bool isALeaf() const;
  /**
    this function with the isALeaf one can help to find if a 
    EMClusterMember is in reality a EMTrackOnEcal 
   It returns 0.0 */  
  virtual Energy_t  maxEnergy() const;
  //--
  ///  Dump for debugging
  virtual void dump() const;
  //--
  ///  return the HitCell pointed
  EMHitCell* whichCell() const { return where;};

  /**
     Return the cell in which the track has been extrapolate.
     Warning: You should call this member function only after that
              the extrapolate() one has been called
     Return a flag (value = -1) to signal if the extrapolation point 
     is outside ECAL or in a HOLE. */
  CellIdx_t whichCellIdx() const ;
  //@}



  /**@name Beheavioral methods */
  //@{
  /**
   Analysis and traversing of the cluster tree is done by ClusterAnalyzers
   that are object with definite analyzing properties. 
   EMClusterMembers MUST ignore these properties: in this way a loose coupling
   between EMClusterMembers and EAClusterAnalyzers is achieved and both
   can vary independently during program evolution.
   Examples of EAClusterAnalyzers are those that should return
   total energy of a cluster, the Center of Gravity, the Number of Cells. */ 
  virtual void analyzeMe(EAClusterAnalyzer & analyzer) const;
   /**
   Analysis and traversing of the cluster tree is done by ClusterAnalyzers
   that are object with definite analyzing properties. 
   EMClusterMembers MUST ignore these properties: in this way a loose coupling
   between EMClusterMembers and EAClusterAnalyzers is achieved and both
   can vary independently during program evolution.
   Examples of EAClusterAnalyzers are those that should return
   total energy of a cluster, the Center of Gravity, the Number of Cells. */ 
  virtual void modifyMe(EDClusterModifier & modifier);
 
  /**
    Fill the internal data with the passed structure.
    Normally the passed data refers to the last hit of the track. */
  void fill(TrackInfo &trackInfo);
  /**
    Extrapolate the hit to the calorimeter. Return a flag to signal
    if the hit is in a cluster or not (maybe outside ECAL or in a HOLE or
    in a cell without signal). */
  EMHitCell* extrapolate( EMClusterTree * tree); 
  /**
    Extrapolate the hit to the calorimeter. Return a flag to signal
    if the hit is in a cluster or not (maybe outside ECAL or in a HOLE or
    in a cell without signal). */
  EMHitCell* extrapolate(Coord_t zCal, EMClusterTree * tree);

  ///  append this track to the  given cluster
  EMClusterMember* appendTo(EMCluster* parent); 

  //--
  //.  evaluate the distance of the extrapolated track from center of cell
  float distancesq(const EMHitCell* hitCell) const;
  //--
  ///  return a constant reference to infos stored
# ifdef CMATCH
   TrackInfo * Infos() const{ return ptOnEcal;};
# else
  const TrackInfo &   Infos() const{ return ptOnEcal;};
# endif
  //@}
private:
  //--
  ///  The extrapolated informations:
# ifdef CMATCH
    TrackInfo*   ptOnEcal;
# else
    TrackInfo   ptOnEcal;
# endif
  //--
  ///  The HitCell in which it is
  EMHitCell * where;

};

typedef EMTrackOnEcal TrackOnEcal;
#endif // EMTRACKONECAL_HH_
