/* see horizontal_rows_in_HPT_Setup.hh for details */

// Author V.Balagura (balagura@mail.desy.de)   (21.02.99).

// Correction made by V.Balagura (balagura@mail.desy.de) (27.03.00):
// changed
//   set<double> y_corners;
// in the beginning of the program to
//   set<double, LessDouble> y_corners(LessDouble(0.01));
// otherwise the rows could be duplicated.
// 

#include "ptset/horizontal_rows_in_HPT_Setup.hh"
#include "ptset/global_error.hh"
#include "ptset/LessDouble.hh"
#include <math.h>
#include <set>

void HorizontalRowsIn_HPT_Chamber::create(HPT_Chamber& chamber) {
  //  cout << "HorizontalRowsIn_HPT_Chamber::create started" << endl;

  // first find the y coordinates of all cell corners 
  // (i.e. in other words project the chamber onto Y axis)
  set<double, LessDouble> y_corners(LessDouble(0.01));
  for (HPT_Chamber::ISection isect=chamber.begin(); isect!=chamber.end();
       ++isect) {
    Grid& grid = (*isect).grid;
    for (HPT_Section::ICell ic=(*isect).begin(); ic!=(*isect).end(); 
	 ++ic) {
      int iy = (*ic).first.iy;
      y_corners.insert(  iy   *grid.y + grid.base_offset.y );
      y_corners.insert( (iy+1)*grid.y + grid.base_offset.y );
    }
  }
  // knowing the y boundaries of horizontal rows calculate
  // the y of their centers

  // the following (yx_centers) will contain:
  // yx_centers[y_of_center_of_every_possible_horizontal_row] = x_map
  // where x_map[center_cell_x] = cell's_ISection_Cell.
  // Here all cells are included which have the y-boundaries
  // y_low < horizontal_row_center_y < y_high
  // In other words: the cells which intersect with the given 
  // horizontal row are sorted by their x_centers and put into the
  // inner map of yx_centers structure.
  // Note that one cell can give its ISection_Cell to many
  // horizontal_rows x_map's (as it e.g. in the HPT chamber corners).

  map<double, 
    map<double, HPT_Setup::ISection_Cell, LessDouble>, LessDouble>
    yx_centers(LessDouble(0.01));
  set<double, LessDouble>::iterator iset, iset_end = y_corners.end();
  --iset_end;
  // calculate the y_center of horizontal rows as 1/2(y_i + y_i+1)
  for (iset=y_corners.begin(); iset!=iset_end;)
    yx_centers.insert
      (make_pair( (*iset + *(++iset))/2., 
		  map<double, HPT_Setup::ISection_Cell, LessDouble>
		  (LessDouble(0.01))));
  // loop over cells
  for (HPT_Chamber::ISection isect=chamber.begin(); isect!=chamber.end();
       ++isect) {
    Grid& grid = (*isect).grid;
    for (HPT_Section::ICell ic=(*isect).begin(); ic!=(*isect).end(); 
	 ++ic) {
      int iy = (*ic).first.iy;
      double 
	y0 =   iy   *grid.y + grid.base_offset.y,
	y1 =  (iy+1)*grid.y + grid.base_offset.y;
      map<double, map<double, HPT_Setup::ISection_Cell, LessDouble>, 
	LessDouble>::iterator iyx;
      double x = ((*ic).first.ix + 0.5)*grid.x + grid.base_offset.x;
      // find the horizontal rows corresponding to this cell, i.e.
      // loop over all rows with the centers 
      // cell_y_low < y_horizontal_row_center < cell_y_high
      for ( iyx = yx_centers.upper_bound(y0);
	    iyx!=yx_centers.end() && (*iyx).first<y1;
	    ++iyx) 
	(*iyx).second[x] = HPT_Chamber::ISection_Cell(isect, ic);
      // put pair(cell_center_x, cell's_ISection_Cell) into the inner
      // yx_corners map
    }
  }
  //  cout << "#rows = " << yx_centers.size() << endl;

  // a few tricks to adjust to the numbering of horizontal rows
  // and the x-numbering of cells in horizontal rows.
  // Aim: horizontal rows are numbered like:
  // -IY, -IY+1, ..., -2, -1,   1, 2, IY-1, IY
  // note that there is no 0-th row. This is done to have the
  // symmetry between y<0 and y>0 halves.

  // X-numbering: the boundary cell which is closest to the beam
  // has #0, others has numbers in ascending order. 
  // Thus for x>0 half: 0, 1, 2, ..., IX (ascending to the right)
  //      for x<0     : IX, ..., 2, 1, 0 (ascending to the left)
  // To implement it: 
  // if abs(x) for the leftmost cell < abs(x) for the rightmost =>
  // this is the x>0 half so go from begin to end (0, 1, 2, ...);
  // otherwise: from end to begin (i.e. from rbegin() to rend()).

  if ( yx_centers.size() % 2 != 0 ) {
    global_error()("number of horizontal rows is odd");
    return;
  }
  // numbering -IY, -IY+1, ..., -2, -1,   1, 2, IY-1, IY is Ok if 
  // # rows is even

  // now convert yx_corners into "cell" data member
  int ix, iy = - int(yx_centers.size()) / 2;
  // simple iy = - yx_centers.size() / 2; is not working since 
  // yx_centers.size() returns unsigned int
  //  cout << "iy_start = " << iy << endl;
  map<double, map<double, HPT_Setup::ISection_Cell, LessDouble>, 
    LessDouble>::iterator iyx;
  for (iyx=yx_centers.begin(); iyx!=yx_centers.end(); ++iyx) {
    map<double, HPT_Setup::ISection_Cell, LessDouble>&
      xmap = (*iyx).second;
    ix = 0;
    // distinguish cases whether to go from begin to end or 
    // vice versa
    if ( fabs((*xmap.begin()).first) < fabs((*xmap.rbegin()).first) ) {
      map<double, HPT_Setup::ISection_Cell, LessDouble>::iterator iiyx;
      for (iiyx=xmap.begin(); iiyx!=xmap.end(); ++iiyx, ++ix)
	cell[Index(ix, iy)] = (*iiyx).second;
    }
    else {
      map<double, HPT_Setup::ISection_Cell, LessDouble>::reverse_iterator
	iiyx;
      for (iiyx=xmap.rbegin(); iiyx!=xmap.rend(); ++iiyx, ++ix) {
	cell[Index(ix, iy)] = (*iiyx).second;
	//	cout << "index included: " << Index(ix, iy) << endl;
      }      
    }
    //    cout << "****************************** iy = " << iy << endl;
    ++iy;
    if ( iy == 0 ) ++iy; // skip iy==0 case
  }
  /*
  cout << "dump" << endl;
  for (map<Index, HPT_Chamber::ISection_Cell, less<Index> >::iterator
	 iter=cell.begin(); iter!=cell.end(); ++iter) {
    cout << (*iter).first     
	 << " section #"
	 << (*iter).second.isection - chamber.begin() 
	 << ", cell " << (*(*iter).second.icell).first << endl;
  }
  cout << "HorizontalRowsIn_HPT_Chamber::create finished" << endl;
  */
}

void HorizontalRowsIn_HPT_Setup::create(HPT_Setup& setup) {
  for (HPT_Setup::IChamber ich=setup.begin(); ich!=setup.end();
       ++ich) {
    (*this)[ich].create((*ich).second);
    if ( global_error() ) break;
  }
}
