/* HPT_CableMakerBasedOnHorizontalRows class
   -----------------------------------------
   intended to define cables for HPT setup.
   It has operator()(istream* is, HPT_Setup::IChamber ichamber); 
   which can be used to construct "interactive" command with one
   parameter.
   E.g. 
     Detector det;
     DetectorMenu menu(&det);
     Cables<HPT_Setup::IChamber> cables;

     menu.hpt.chamber("define_cable") = make_interactive_command_1par
       (HPT_CableMakerBasedOnHorizontalRows(&cables),
         (HPT_Setup::IChamber*)0);
   Here "menu.hpt" is the menu for High-Pt Setup, 
   "menu.hpt.chamber" is the menu for chambers in High-Pt Setup.
   (see file ptset/detector_menu.hh).

   The corresponding interactive command "define_cable" is positioned
   inside the chamber menu. Thus there is no need to write the chamber
   name in defining the cable.
   The format is:
   /sub-det{hpt_inner
        /chamber{chamber_name
	     /define_cable{ PT1  [0:16, 5) skip 5 [16:32, 5) ... }
   
   Here PT1 is the cable name (in Cables<HPT_Setup::IChamber>& cables;),
   [0:16, 5), [16:32, 5) are IndexRange_X (defined in index_range.hh),
   skip 5 is a SkipRange (defined in skip_range.hh).
   I.e. cable can contain "empty" regions (use "skip" to define them),
   the number of IndexRange_X's and SkipRange's are not limited.

   IndexRange is defined using the following conventions:
   Y-numbering:
   -IY, -IY+1, ..., -2, -1,   1, 2, IY-1, IY
   note that there is no 0-th row. This is done to have the
   symmetry between y<0 and y>0 halves.

   X-numbering: the boundary cell which is closest to the beam
   has #0, others has numbers in ascending order. 
   Thus for x>0 half: 0, 1, 2, ..., IX (ascending to the right)
        for x<0     : IX, ..., 2, 1, 0 (ascending to the left)
   
   Horizontal rows are defined in this way using 
   HorizontalRowsIn_HPT_Setup rows;
   (defined in horizontal_rows_in_HPT_Setup.hh).
   Its usage: rows[Index(ix, iy)] -> ISection_Cell.
   */

// Author V.Balagura (balagura@mail.desy.de)   (21.02.99).

#ifndef _HPT_cable_maker_based_on_horizontal_rows_hh
#define _HPT_cable_maker_based_on_horizontal_rows_hh 1

#include "ptset/horizontal_rows_in_HPT_Setup.hh"
#include "ptset/cables.hh"

struct HPT_CableMakerBasedOnHorizontalRows {
  Cables<HPT_Setup::IChamber>& cables;
  HorizontalRowsIn_HPT_Setup rows;

  HPT_CableMakerBasedOnHorizontalRows
  (Cables<HPT_Setup::IChamber>* HPT_cables) : cables(*HPT_cables) {} 
  // check global_error() after it
   
  void operator()(istream* is, HPT_Setup::IChamber ichamber);

private:
  void fill(string* rest, Cable<HPT_Setup::IChamber>& cable,
	    HorizontalRowsIn_HPT_Chamber& rows_in_chamber,
	    HPT_Chamber::ISection dummy_is, bool is_first);
};

#endif
