/* InteractiveParser is the general class which is used to strip
   the input to "commands" and "arguments". 
   struct InteractiveParser {
     const string EOF_MESSAGE = "end of file reached";
     struct Output {
       list<string>* commands;
       istream*      arguments;
     };
     virtual Output get_input(istream* is) = 0;
   };
   First call get_input(istream* is); the result is of type 
   InteractiveParser::Output - it contains "commands" and "arguments"
   which can be used as input for InteractiveMenu.

   The derived classes are:
   ComArg_InteractiveParser (abstract class) - it holds
     list<string> coms;
     strstream    args;
   in itself, returns pointers to them.
   InteractiveParser_WithDelimiters (also abstract) - strips the input
   using 
   char get_coms(istream* is, char com_delimiter, 
		bool (*arg_delimiter)(char));
   in the following way: everything before com_delimiter is a "command",
   everything next up to the first symbol c which satisfy
   (*arg_delimiter)(c) == true; is an "argument".

   SlashLine_InteractiveParser - can be used directly (not abstract):
   delimiter between "com" and "args" is whitespace,
   end of "args" mark: new line symbol. This class can be used to get 
   one line command.

  SlashBlock_InteractiveParser:
  delimiter between "com" and "args" is '{', 
   end of "args" mark: **corresponding** closed bracket '}'. 
   This class can be used to strip the input which is represented as 
   "blocks" of information (probably nested). 
*/

// Author V.Balagura (balagura@mail.desy.de)   (24.04.98).

#ifndef _interactive_parser_hh
#define _interactive_parser_hh 1

#include <string>
#include <list>
#include <strstream.h>
#include "ptset/global_error.hh"
#include <ctype.h> // for int isspace(char)

struct InteractiveParser {
 public:
  static string EOF_MESSAGE;
  struct Output {
    list<string>* commands;
    istream*      arguments;
    Output() {}
    Output(list<string>* coms, istream* args)
      : commands(coms), arguments(args) {}
  };
  virtual Output get_input(istream* is) = 0;
  virtual InteractiveParser* clone() const = 0;
  virtual ~InteractiveParser() {}
};

class ComArg_InteractiveParser : public InteractiveParser {
 protected:
  list<string> coms;
  strstream    args;

  void clear_output();
  void print_error(const string& message, string* add);
  virtual void fill_output(istream* is) = 0;
 public:
  virtual Output get_input(istream* is) { 
    fill_output(is);
    return Output(&coms, &args); 
  }
};
class  InteractiveParser_WithDelimiters : public ComArg_InteractiveParser {
 protected:
  char get_coms(istream* is, char com_delimiter, 
		bool (*arg_delimiter)(char));
  // returns character last read
};

class  SlashLine_InteractiveParser : public InteractiveParser_WithDelimiters {
 protected:
  static bool is_space(char c) { return isspace(c); }
  virtual void fill_output(istream* is);
 public:
  virtual InteractiveParser* clone() const {
    /* there is no copy constructor for strstream class.
       Therefore SlashLine_InteractiveParser also does not have copy
       constructor => here default constructor is used */
    return new SlashLine_InteractiveParser();
  }
};
class  SlashBlock_InteractiveParser
: public InteractiveParser_WithDelimiters {
 protected:
  static bool is_left_bracket(char c) { return c == '{'; }
  virtual void fill_output(istream* is);
 public:
  virtual InteractiveParser* clone() const {
    /* there is no copy constructor for strstream class.
       Therefore SlashBLock_InteractiveParser also does not have copy
       constructor => here default constructor is used */
    return new SlashBlock_InteractiveParser();
  }
};

#endif
