
// ********************************************************************
//
//  Routine:      rc99AStandAlone
//  Author:       T.Oest/D.Samtleben
//
//  Description:  stand alone reconstruction
//
//                reco level 0: no reconstruction
//                           1: hitb only
//                           2: reconstruction only
//                           3: hitb + reconstruction
//
//  Modification:  
//      - drop RICH hits
//      - recoTARGET,... variables no longer static
//                                                        27/05/99 Oest
//      - VDS==SVD for kumac-steering
//      - error-message if given subdetector is not known
//                                                        10/06/99 Samtleben
//      - if at least one reconstruction package of a subdetector
//        is called the Rxxx-tables are cleared
//                                                        01/07/99 Samtleben
//      - bug fix: proper termination of input[]
//      - include CATS: recoVDSmodule
//      - correct calling sequenz for rc99ATARGET
//                                                        26/07/99 Oest
//     -- global alignment transformation added 
//                                                        06/09/99 T.Oest
//     -- RICH hits are no longer removed since they are needed for
//        the particle ID                                 14/09/99 T.Oest 
//     -- move VDS global alignment transformation to rc99AVDS
//                                                        20/09/99 T.Oest
//     -- filling of GTAR bank moved to acReadTARGET      24/09/99 T.Oest
//     -- include global VDS alignment transformation
//                                                        29/10/99 T.Oest
//     -- ITR added                                       07/02/00 T.Oest
//     -- use ARTE_mark()                                 12/05/00 AG
//     -- introduce timing                                17/05/00 AG
//     -- timing info into globals                        18/05/00 AG
//     -- creation of chrono jobs only once per job       24/05/00 D.Samtleben
//
// ********************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "reco/rc99Areco.hh"
#include "reco/rcflags.hh"

// globals
float time_rc99AVDS  = -1;
float time_rc99AITR  = -1;
float time_rc99AHPT  = -1;
float time_rc99AOTR  = -1;
float time_rc99ARICH = -1;
float time_rc99ATRD  = -1;
float time_rc99AECAL = -1;
float time_rc99AMUON = -1;

// prototypes
void acRSEGinGlobalCoordinates(const int CMPnumber);

//********************************************************************
// rc99AStandAlone
//********************************************************************
int rc99AStandAlone() {

//===================================================================
// remove the Rxxx tables (for real data)
//===================================================================

//------  remove Rxxx-tables
  if(recoVDS||recoECAL||recoRICH||recoHPT||recoTRD||recoMUON||
     recoOTR){
    ArteTable<RREL>::clear();
    ArteTable<RPNT>::clear();
    ArteTable<RVER>::clear();
    ArteTable<RHIT>::clear();
    ArteTable<RSEG>::clear();
    ArteTable<RTRA>::clear();
    ArteTable<RCCL>::clear();
  }
//===================================================================
//  stand alone tracking
//===================================================================

  // Timers
  static Chrono *cron_rc99AVDS  = 0;
  static Chrono *cron_rc99AITR  = 0;
  static Chrono *cron_rc99AHPT  = 0;
  static Chrono *cron_rc99AOTR  = 0;
  static Chrono *cron_rc99ARICH = 0;
  static Chrono *cron_rc99ATRD  = 0;
  static Chrono *cron_rc99AECAL = 0;
  static Chrono *cron_rc99AMUON = 0;

  static bool first = 1;
  if( first ){
    first = 0;
    ChronoManager CM;
    CM.instance();
    
    cron_rc99AVDS  = CM.createChrono( "cron_rc99AVDS"  );
    cron_rc99AITR  = CM.createChrono( "cron_rc99AITR"  );
    cron_rc99AHPT  = CM.createChrono( "cron_rc99AHPT"  );
    cron_rc99AOTR  = CM.createChrono( "cron_rc99AOTR"  );
    cron_rc99ARICH = CM.createChrono( "cron_rc99ARICH" );
    cron_rc99ATRD  = CM.createChrono( "cron_rc99ATRD"  );
    cron_rc99AECAL = CM.createChrono( "cron_rc99AECAL" );
    cron_rc99AMUON = CM.createChrono( "cron_rc99AMUON" );
  }

  // VDS
  if (recoVDS){
    ARTE_mark( "rc99AVDS", 0 );
    cron_rc99AVDS->start();
    int SVDflag = rc99AVDS(recoVDS);
    cron_rc99AVDS->stop();
    time_rc99AVDS = cron_rc99AVDS->lastTime();
  }

  // ITR
  if (recoITR){
    ARTE_mark( "rc99AITR", 0 );
    cron_rc99AITR->start();
    int ITRflag = rc99AITR(recoITR);
    cron_rc99AITR->stop();
    time_rc99AITR  = cron_rc99AITR->lastTime();
  }

  // HPT
  if (recoHPT){
    ARTE_mark( "rc99AHPT", 0 );
    cron_rc99AHPT->start();
    int HPTflag = rc99AHPT(recoHPT);
    cron_rc99AHPT->stop();
    time_rc99AHPT = cron_rc99AHPT->lastTime();
  }

  // OTR
  if (recoOTR){
    ARTE_mark( "rc99AOTR", 0 );
    cron_rc99AOTR->start();
    int OTRflag = rc99AOTR(recoOTR);
    cron_rc99AOTR->stop();
    time_rc99AOTR = cron_rc99AOTR->lastTime();
  }

  // RICH
  if (recoRICH){
    ARTE_mark( "rc99ARICH", 0 );
    cron_rc99ARICH->start();
    int RICHflag = rc99ARICH(recoRICH);
    cron_rc99ARICH->stop();
    time_rc99ARICH = cron_rc99ARICH->lastTime();
  }

  // TRD
  if (recoTRD){
    ARTE_mark( "rc99ATRD", 0 );
    cron_rc99ATRD->start();
    int TRDflag = rc99ATRD(recoTRD);
    cron_rc99ATRD->stop();
    time_rc99ATRD = cron_rc99ATRD->lastTime();
  }

  // MUON
  if (recoMUON){
    ARTE_mark( "rc99AMUON", 0 );
    cron_rc99AMUON->start();
    int MUONflag = rc99AMUON(recoMUON);
    cron_rc99AMUON->stop();
    time_rc99AMUON = cron_rc99AMUON->lastTime();
  }

//===================================================================
//  transformation of segments/RHIT to global coordinates
//===================================================================
  
  ARTE_mark( "acRSEGinGlobalCoordinates", 0 );
  if (recoVDSmodule & 1) acRSEGinGlobalCoordinates(1);
  if (recoVDSmodule & 2) acRSEGinGlobalCoordinates(65537);
  if (recoRICH)          acRSEGinGlobalCoordinates(67108864);
  
//===================================================================
//  ECAL needs tracks for the 2. reconstruction step
//===================================================================

  // ECAL
  if (recoECAL){
    ARTE_mark( "rc99AECAL", 0 );
    cron_rc99AECAL->start();
    int ECALflag = rc99AECAL(recoECAL);
    cron_rc99AECAL->stop();
    time_rc99AECAL = cron_rc99AECAL->lastTime();
  }

//===================================================================
//  flag for stand alone reconstruction
//===================================================================

  return 0;  
}

