//--*-C++-*--
#ifndef LAMBDAPHI_HH
#define LAMBDAPHI_HH

#include "rich/ricutil.h"


// ------------------------------------------------------------------
// ---  Angle corrections
// ------------------------------------------------------------------

/** Description of RICH angular correction class */
class Corrections {
private:
  float phi,lam,rad,Xt,Yt,dlam;   // particle geometry

  float s,lam0,th0,R,z,zz,zzz,T;  // RICH geometry

  float verticalD_u,horizontalD_u;    // distortion
  float verticalP_u,horizontalP_u;    // position
  float Avh_u;                        // v-h coupling
  float verticalD_d,horizontalD_d;    // distortion
  float verticalP_d,horizontalP_d;    // position
  float Avh_d;                        // v-h coupling
  float h,Hlam,Hphi,ddaa_u,ddaa_d;    // helpers

  void  calcH();
  float calcA();
  float verticalDistortion  ();
  float horizontalDistortion(); 
  float verticalPosition    ();
  float horizontalPosition  ();

public:
  /// Constructor
  Corrections();
  /// Constructor
  Corrections(float r0,float s0,float t0,float l0,float t0,float z0);

  /// Initialization of photon corrections for a ring in stand-alone ringsearch
  void   newRing(float phi,float lam,float rad);
  /// Initialization pf photon corrections for a ring in track-seeded mode
  void   newRing(float phi,float lam,float rad,float Xt,float Yt);

  inline float &upperDv() { return verticalD_u;   }
  inline float &upperDh() { return horizontalD_u; }
  inline float &upperPv() { return verticalP_u;   }
  inline float &upperPh() { return horizontalP_u; }
  inline float &upperA()  { return Avh_u;         }
  inline float &lowerDv() { return verticalD_d;   }
  inline float &lowerDh() { return horizontalD_d; }
  inline float &lowerPv() { return verticalP_d;   }
  inline float &lowerPh() { return horizontalP_d; }
  inline float &lowerA()  { return Avh_d;         }


  inline float verticalShift(int plane) { 
    if (plane) return verticalP_d;
    return verticalP_u;
  }

  inline float horizontalShift(int plane) {
    if (plane) return horizontalP_d;
    return horizontalP_u;
  }

  inline float verticalDistortion(int plane) {
    if (plane) return verticalD_d;
    return verticalD_u;
  }

  inline float horizontalDistortion(int plane) {
    if (plane) return horizontalD_d;
    return horizontalD_u;
  }

  inline float skewing(int plane) {
    if (plane) return Avh_d;
    return Avh_u; 
  }

  inline float radius(float p,float l,int plane) {
    float d_phi=dPhi   (p,l,plane);
    float d_lam=dLambda(p,l,plane);
    return sqrt(d_phi*d_phi+d_lam*d_lam);
  }
 

  /// Corrected horizontal angle (photon-ring center)
  inline float dPhi(float p,float l,int plane) {
    if (plane)
      return  ( (p-phi+horizontalP_d)*verticalD_d - (l-lam+verticalP_d)*Avh_d ) * ddaa_d; 
    return    ( (p-phi+horizontalP_u)*verticalD_u - (l-lam+verticalP_u)*Avh_u ) * ddaa_u;
  }
  
  /// Corrected vertical angle (photon-ring center)
  inline float dLambda(float p,float l,int plane) {
    if (plane) 
      return   ( (l-lam+verticalP_d)*horizontalD_d - (p-phi+horizontalP_d)*Avh_d ) * ddaa_d;
    return     ( (l-lam+verticalP_u)*horizontalD_u - (p-phi+horizontalP_u)*Avh_u ) * ddaa_u;
  }



  /*
  float dPhi(float p,float l,int plane) {
    if (plane)  return (p-phi+horizontalP_d)/horizontalD_d;
    return (p-phi+horizontalP_u)/horizontalD_u;
  }

  float dLambda(float p,float l,int plane) {
    if (plane) 
      return (l-lam+verticalP_d)/verticalD_d;
    return   (l-lam+verticalP_u)*verticalD_u;
  }
  */
};







// ------------------------------------------------------------------
// ---  RICH lambda/phi angles
// ------------------------------------------------------------------


/** Description of RICH reconstruction space */
class LambdaPhi {
private:
  Vec3   Vertex;
  float  *pphi,*llam;
      
  double  d1,d2,d12;
  double  d1d1,d2d2,d12d12;

  inline  double  a1(double u) {
    return sqrt(1-d2d2+u*u)-sqrt(d1d1-d2d2+u*u);
    }
  inline  double  a2(double u) {
    return sqrt(1-d2d2+u*u)+u;
    }
  inline  double  f (double u) {
    double A1=a1(u);
    double A2=a2(u);
    return .5+(d12d12-A1*A1-A2*A2)/(4*A1*A2)-d2d2+u*u;
    }
  inline  double df(double u) {
    double du=u/100.;
    if (du<=0) du=1e-3;
    return (f(u+du)-f(u))/du;
    }

  Corrections *corr;

public:
  /// Constructor
  LambdaPhi(Vec3 Vertex=Vec3(0,0,0));
  /// Returns phi coordinate of a channel
  inline float phi(int i)    const { return pphi[i]; }
  /// Returns lambda coordinate of a channel
  inline float lambda(int i) const { return llam[i]; }
  /// Solves SM reflexion
  Vec3         findReflectedDirection(Vec3 source, Vec3 destination, double R, Vec3 coc);
  /// Returns phi coordinate of a vector
  inline float phi(Vec3 n)    const { return atan(n.x()/n.z()); }
  /// Returns lambda coordinate of a vector
  inline float lambda(Vec3 n) const { return atan(n.y()/n.z()*cos(phi(n))); }
  /// Returns phi coordinate from tx, ty
  inline float phi(float tx, float ty)     const { return atan(tx); }
  /// Returns lambda coordinate from tx, ty
  inline float lambda(float tx, float ty)  const { return atan(ty*cos(phi(tx,ty))); }
  /// Returns tx coordinate from phi, lambda
  inline float tx(float phii, float lamm)  const { return tan(phii); }
  /// Returns ty coordinate from phi, lambda
  inline float ty(float phii, float lamm)  const { return tan(lamm)/cos(phii); }

  /// Link to correction class
  inline Corrections *Correct() { return corr; }
};

/** Global variable for RICH reconstruction space. */
extern LambdaPhi *RICH_LambdaPhi;





#endif

