//-----------------------------------------------------------------------//
//
//  RingSearchARTE: ARTE interface for stand-alone ring reconstruction
//
//  Author: Reinhard Eckmann/Denis Dujmic
//
//  Created: May-98
//
//  Modifications:
//    May 15, 2000 (DD) nhit filled with Ring::numhit_3sigma
//    May 24, 2000 (DD) OTR based reconstruction: ringsearchRSEG()
//    May 29, 2000 (DD) ringsearchRSEG: use underdogs, SN>4 with seed,
//                      RingSearchAttrib::finer_search_step added
//    May 31, 2000 (DD) makeHitMap: check if radius list empty
//
//-----------------------------------------------------------------------//
#include "rich/rich.h"
#include "rise/RingSearchARTE.h"
#include "rise/RingSearchTiming.h"
#include "arte/RSEG_cmp.hh"



#ifdef RISE_ROOT
RingSearchTiming *RICH_SearchTiming=0; 
#endif

/////////////////////////// Ring search ////////////////////////////////////////

RingSearchARTE::RingSearchARTE() : RingSearchDB() {

#ifdef RISE_TIMING
  if (!RICH_SearchTiming) RICH_SearchTiming=new RingSearchTiming;
#endif

  if (RichRun::magnet) printf(" $$$ RISE Inform: RingSearch with magnet ON\n");
  else printf(" $$$ RISE Inform: RingSearch with magnet OFF\n");

  n_mass=5;
}


int RingSearchARTE::makeHitMap() {

  int verb=0;

  // -- reset hit bank
  h.reset();
  d.nextevent();

  // -- ocupancy cut
  int n0=RichRun::gede->hit1.index();
  int n1=RichRun::gede->hitl.index();
  if (n1-n0>occupancyCut()) return 1;
  if (n1-n0<=0)             return 2;

  // -- track based ring search
  //float min_phi=0, max_phi=0;
  //float min_lam=0, max_lam=0;
  float min_rad=0, max_rad=0;
  if ( RingSearchTrack::exist && RingSearchAttrib::id) {
    if (verb>1) cout<<"new ring input:"<<RingSearchTrack::phi<<" "<<RingSearchTrack::lambda<<" "<<RingSearchTrack::bx
		  <<" ("<<-RingSearchTrack::phi*4.5<<") "
		  <<RingSearchTrack::by<<endl;
    float rr=ArtePointer<RITH>(1)->rad0;
    RICH_LambdaPhi->Correct()->newRing(RingSearchTrack::phi, RingSearchTrack::lambda, rr,
				       RingSearchTrack::bx,RingSearchTrack::by);
    //min_phi = minSearchRadius()*.001 - RingSearchTrack::max_phi_residual;
    //max_phi = maxSearchRadius()*.001 + RingSearchTrack::max_phi_residual;
    //min_lam = minSearchRadius()*.001 - RingSearchTrack::max_lambda_residual;
    //max_lam = maxSearchRadius()*.001 + RingSearchTrack::max_lambda_residual;
    //min_rad = minSearchRadius()*.001 - .5*(RingSearchTrack::max_phi_residual + RingSearchTrack::max_lambda_residual);
    //max_rad = maxSearchRadius()*.001 + .5*(RingSearchTrack::max_phi_residual + RingSearchTrack::max_lambda_residual);
    //min_rad =  minSearchRadius()*.001 - RingSearchTrack::max_lambda_residual;
    //max_rad =  maxSearchRadius()*.001 + RingSearchTrack::max_lambda_residual;
    min_rad =  60;
    max_rad =  0;
    if (!radius_list.size()) {
      cout<<" $ RICH: RingSearchARTE::makeHitMap: search radius list empty "<<endl;
      return 3;
    }
    for (vector<float>::iterator irad=radius_list.begin(); 
	 irad!=radius_list.end(); ++irad) {
      if (min_rad>(*irad)) min_rad=(*irad);
      if (max_rad<(*irad)) max_rad=(*irad);
    }
    min_rad*=.001;
    max_rad*=.001;
    min_rad-=RingSearchTrack::max_lambda_residual;
    max_rad+=RingSearchTrack::max_lambda_residual;
    if (verb) cout<<"min/max="<<min_rad<<"/"<<max_rad<<endl;
  }


  // -- append photons
  for (int i=n0;i<n1;i++) {

    ArtePointer<HITB> hitb(i);

    // -- check if photon has been used
    if (!RingSearchAttrib::reuse &&
	hitb->rseg.begin()!=hitb->rseg.end()) continue;    

    ArteRelationIterator<DIGB> idigb=hitb->digb.begin();
    ArtePointer<DIGB> digb(idigb);
    int id    = digb->id;
    int plane = (smFromChannelID(id)>6) ? 1 : 0 ;
    if (!RICH_Detector->ChannelExist(id)) {
      static int ierr=0;
      if (++ierr<10)
	cout<<" $$$ RICH: requested channel not initialized - check DB/data files!"<<endl;
      continue;
    }
    int size  = RICH_Detector->pm().numchannel();
    int imtra = 9999;

    // -- track seed search
    if (RingSearchTrack::exist) {
      //cout<<"RingSearchTrack::exist"<<endl;
      float dx=fabs(RICH_LambdaPhi->Correct()->dPhi   (hitb->q, hitb->r, plane));
      float dy=fabs(RICH_LambdaPhi->Correct()->dLambda(hitb->q, hitb->r, plane));
      //if (dx>max_phi || dx<min_phi) continue;
      //if (dy>max_lam || dy<min_lam) continue;
      float dr=sqrt(dx*dx+dy*dy);
      if (dr>max_rad || dr<min_rad) continue;
    }
    if (ArteTable<MTRA>::size()) { // MC mode
      ArteRelationIterator<MIMP> imimp=hitb->mimp;
      ArtePointer<MTRA> mtra=(*imimp)->mtra;
      imtra = mtra.index();
    }
    h.append(d.dualx(hitb->q*milliradian),d.dualy(hitb->r*milliradian), plane, size, imtra, hitb);
    if (verb) cout<<".";
  }
  if (verb) cout<<" numhit="<<h.numhit<<endl;
  if (ArteTable<MTRA>::size()) h.randomize(); 
  return 0;
}


void RingSearchARTE::makeRadiusList(float r1, float r2, float drad) {
  int verb=0;
  
  if (verb) cout<<"RingSearchARTE::makeRadiusList"<<endl;

  if (radius_list.size()) radius_list.clear();

  if (RingSearchTrack::exist && RingSearchAttrib::id) {
    float oldrad=100;

    for (int i=0; i<5; i++) {
      float rr=R2P2rad2[i]*1e6;
      float R2P2sigma=1e6/sqrt(iR2P2sigma2[i]);
      if (rr<0) continue;
      float r=sqrt(rr);
      if (RingSearchAttrib::finer_search_step) { 
	float dr=sqrt(rr+4*R2P2sigma)-r;
	if (r<minSearchRadius()) continue;
	if (drad<0) {
	  if (oldrad-r>2*dr) radius_list.push_back(r);
	  if (oldrad-r>3*dr) radius_list.push_back(r+dr);
	  if (oldrad-r>dr) { radius_list.push_back(r-dr); oldrad=r; }
	}
	else {
	  if (oldrad-r>3*dr)   radius_list.insert(radius_list.begin(),r+dr);
	  if (oldrad-r>2*dr)   radius_list.insert(radius_list.begin(),r);
	  if (oldrad-r>dr)   { radius_list.insert(radius_list.begin(),r-dr); oldrad=r; }
	}
      }
      else radius_list.push_back(r);
    }

  }
  else {
    if (drad<0) for (float radius=r2;radius>r1;radius+=drad) {// mrad /////
      radius_list.push_back(radius);
    }
    else for (float radius=r1;radius<r2;radius+=drad) {// mrad /////
      radius_list.push_back(radius);
    }
  }

}


void RingSearchARTE::getrings(float r1,float r2,float dr,int t) {
  int verb=0;

#ifdef RISE_TIMING
  RICH_SearchTiming->startTiming();
#endif

  switch (RingSearchAttrib::seed_type) {
  case RingSearchType::rtra_seed: if (!SetSeedRTRA()) return; break;
  case RingSearchType::rseg_seed: if (!SetSeedRSEG()) return; break;
  default: cout<<"RingSearchARTE::getrings: Wrong seed type"<<endl;
  }
  makeRadiusList(r1,r2,dr);
  if (makeHitMap()) return;

  int threshold=t;
  for (vector<float>::iterator irad=radius_list.begin();
       irad!=radius_list.end(); ++irad) {
    //////////if (RingSearchTrack::exist && (*irad)<40) threshold=3;
    //////////else if (RingSearchTrack::exist) threshold=4;
    if (verb) cout<<"radius="<<(*irad)<<"  threshold="<<threshold<<"  seed="<<RingSearchTrack::exist<<endl;
    d.define((*irad)*1e-3*milliradian,threshold);
    d.ringsearch(h);
  }
  
  if (RingSearchAttrib::choose_best_ring) Choose();
  if (RingSearchAttrib::bookrseg)         FillRSEG();

#ifdef RISE_TIMING
  RICH_SearchTiming->stopTiming();
  RICH_SearchTiming->fillTimingHistograms();
#endif
}



void RingSearchARTE::getrings(int i) {
  int verb=0;

  if (R2P2rad2[i]<1e-4) return;

  Ring &testring=getRing(0);

  // -- set parameters
  testring.thetaC()=sqrt(R2P2rad2[i]);
  bool old_use_phi=RingSearchAttrib::use_seed_phi;
  bool old_use_lam=RingSearchAttrib::use_seed_lambda;
  RingSearchAttrib::use_seed_phi=true;
  RingSearchAttrib::use_seed_lambda=true;

  if (!testring.calcring(h)) return;
  else numrings()=1;

  if (RingSearchAttrib::bookrseg) FillRSEG();

  // -- restore parameters
  RingSearchAttrib::use_seed_phi   =old_use_phi;
  RingSearchAttrib::use_seed_lambda=old_use_lam;  
}




void RingSearchARTE::getrings() {
  if (!ArteTable<MTRA>::nrow()) getSearchParameters(ArtePointer<EVHD>(1)->run);
  getrings(minSearchRadius(),maxSearchRadius(),searchStep(),threshold());
}


//////////////////////////  Seed setup //////////////////////////////////

bool RingSearchARTE::SetSeedRTRA() {
  int verb=0;

  if (verb) cout<<"SetSeedRTRA"<<endl;

  if (!RingSearchTrack::exist) return true;

  // -- set track parameters
  RingSearchTrack::ip    = RingSearchTrack::rtra->pf;
  //!!!!!!!!!!!!!!!!! make sure RPNT is right one !!!!!!!!!!!!!!
  ArtePointer<RPNT> rpnt;
  float max_z=0;
  for(ArteRelationIterator<RPNT> irpnt=RingSearchTrack::rtra->rpnt.begin();
      irpnt!=RingSearchTrack::rtra->rpnt.end(); irpnt++) {
    if ((*irpnt)->zf>max_z && 
	(*irpnt)->zf<RingSearchTrack::max_rpnt_z &&
	(*irpnt)->zf>RingSearchTrack::min_rpnt_z) { rpnt=(*irpnt); max_z=(*irpnt)->zf; }
  }
  if (max_z<RingSearchTrack::min_rpnt_z) {
    if (verb) cout<<" $$$ RICH: RingSearchARTE::SetSeed no valid RPNT"<<endl;
    return false;
  }


  // -- get track parameters inside RICH
  float x  = rpnt->xf;
  float y  = rpnt->yf;
  float z  = rpnt->zf;
  float tx = rpnt->txf;
  float ty = rpnt->tyf; 

  RingSearchTrack::phi    = RICH_LambdaPhi->phi   (tx, ty);
  RingSearchTrack::lambda = RICH_LambdaPhi->lambda(tx, ty);
  RingSearchTrack::bx     = ( x - tx*z )*.01; // m
  RingSearchTrack::by     = ( y - ty*z )*.01; // m

  // -- calculate expected positions and errors on R2P2 lines
  if (RingSearchAttrib::id) makeR2P2lines();
  return true;
}

bool RingSearchARTE::SetSeedRSEG() {
  int verb=0;

  if (verb) cout<<"SetSeedRSEG"<<endl;

  if (!RingSearchTrack::exist) return true;


  // -- get track parameters inside RICH
  float x  = RingSearchTrack::rseg->xe;
  float y  = RingSearchTrack::rseg->ye;
  float z  = RingSearchTrack::rseg->ze;
  float tx = RingSearchTrack::rseg->txe;
  float ty = RingSearchTrack::rseg->tye; 

  /*int nt=0;
  for (ArteRelationIterator<RHIT> irhit=RingSearchTrack::rseg->rhit.begin();
       irhit!=RingSearchTrack::rseg->rhit.end(); ++irhit) {
    tx+=(*irhit)->tx;
    ty+=(*irhit)->ty;
    nt++;
    cout<<"rhit="<<(*irhit)->x<<" "<<(*irhit)->y<<" "<<(*irhit)->tx<<" "<<(*irhit)->ty<<endl;
  }
  tx/=nt;
  ty/=nt;
  cout<<"angles:  rseg="<<RingSearchTrack::rseg->txe<<","<<RingSearchTrack::rseg->tye
      <<"  average="<<tx<<" "<<ty
      <<"  first/last="<<(RingSearchTrack::rseg->xe-RingSearchTrack::rseg->xf)/(RingSearchTrack::rseg->ze-RingSearchTrack::rseg->zf)
      <<","<<(RingSearchTrack::rseg->ye-RingSearchTrack::rseg->yf)/(RingSearchTrack::rseg->ze-RingSearchTrack::rseg->zf)
      <<endl;
      */

  // -- set track parameters
  //RingSearchTrack::ip    = RingSearchTrack::rseg->pf;
  RingSearchTrack::ip    = (x-tx*z)/(-288.);

  RingSearchTrack::phi    = RICH_LambdaPhi->phi   (tx, ty);
  RingSearchTrack::lambda = RICH_LambdaPhi->lambda(tx, ty);
  RingSearchTrack::bx     = ( x - tx*z )*.01; // m
  RingSearchTrack::by     = ( y - ty*z )*.01; // m

  // -- calculate expected positions and errors on R2P2 lines
  if (RingSearchAttrib::id) makeR2P2lines();
  return true;
}



void RingSearchARTE::Choose() {
  //cout<<"RingSearchARTE::Choose() numrings="<<numrings()<<endl;

  if (!RingSearchTrack::exist) return;
  // -- in case of seeded search find best ring
  float best_S=0;
  int   besti=-1;
  for (int i=0; i<numrings(); i++) {
    Ring &ring=getRing(i);
    ring.accept=false;
    if (!ring.IsCloseToSeed()) continue;
    // -- find best significance or S/Sqrt(N)
    if (RingSearchAttrib::type==RingSearchType::lkbg) {
      if (best_S<ring.refit_S) {
	best_S=ring.refit_S;
	besti=i;
      }
    }
    else {
      float sn=(ring.numhit_bgrd>0) ? ring.numhit_refit/sqrt(ring.numhit_bgrd) : 100;
      if (best_S<sn) {
	best_S=sn;
	besti=i;
      }
    }
  }
  if (besti<0) return;
  Ring &bestring=getRing(besti);
  bestring.accept=true;
}



extern "C"  void rkmain_(const int    & icovar, 
              const double & zout, 
              const double & fstep, 
              const double * parameters, 
              const double * prec, 
              const double * p_track, 
              const double * derivatives, 
              const int    & ierror);

int RingSearchARTE::swim(const double *source,double *dest,const double zout) {
  const int    icovar=2;
  double       fstep=50;
  double       prec[4];
  double       derivatives[5][5];
  int          ierror;

  prec[0] = 0.001/fabs(zout-source[0]);
  prec[1] = prec[0];
  prec[2] = 0.0001/fabs(zout-source[0]);
  prec[3] = prec[2];

  rkmain_(icovar,zout,fstep,source,prec,dest,&derivatives[0][0],ierror);

  return ierror;
}


////////////////////////// booking of ARTE tables ////////////////////////


void RingSearchARTE::FillRSEG() {
  int verb=0;

  RSEG r;
  ArtePointer<RSEG>rseg;
  //cout<<"RingSearchARTE::FillRSEG() numrings="<<numrings()<<endl;
  for (int i=0; i<numrings(); i++) {

    Ring &ring=getRing(i);

    // -- check distance seed-ring center
    if (RingSearchTrack::exist && 
	RingSearchAttrib::choose_best_ring && 
	(!ring.accept) ) continue;

    // -- ring info
    r.xf = ring.phi();
    r.yf = ring.lambda();

    r.time   = ring.thetaC();
    float r0 = ArtePointer<RITH>(1)->rad0;
    r.dtim   = cos(r0)/cos(r.time);

    r.txf = ring.tx();
    r.tyf = ring.ty();

    r.pf  = float(ring.numHits());
    r.ze  = ring.numhit_refit;
    r.ye  = ring.numhit_bgrd;
    r.xe  = ring.numhit_3sigma;
    r.zf  = ring.f;
    r.txe = ring.Rb;
    r.tye = ring.refit_S;
    r.pe  = ring.sigma_signal;

    r.nhit= (int)ring.numhit_3sigma;

    r.chi = ring.chi2;

    r.cvf[ 0] = 0;
    r.cvf[ 1] = 0;
    r.cvf[ 2] = 0;
    r.cvf[ 5] = ring.covariance(0,0);
    r.cvf[ 8] = ring.covariance(0,1);
    r.cvf[12] = ring.covariance(0,2);
    r.cvf[ 9] = ring.covariance(1,1);
    r.cvf[13] = ring.covariance(1,2);
    r.cvf[14] = ring.covariance(2,2);
    r.cvf[ 3] = 0;
    r.cvf[ 6] = 0;
    r.cvf[10] = 0;
    r.cvf[ 4] = 0;
    r.cvf[ 7] = 0;
    r.cvf[11] = 0;

    // -- component
    r.cmp = Rsegc::rich;

    // -- fit type
    if (RingSearchTrack::exist) r.fit=12;       // seeded track
    else {
      float ratio=ring.thetaC()/r0;
      ratio*=ratio;
      if (ring.numhit_refit>50*ratio) r.fit=10; // possible gamma conversion
      else r.fit=8;                             // charged track found standalone
    }

    // -- book relations
    // -- RSEG
    rseg=ArteTable<RSEG>::push_back(r);

    // -- RSEG-MTRA
    unsigned int iMtra=ring.mctrack();
    if (ArteTable<MTRA>::size() && iMtra<=ArteTable<MTRA>::size() && iMtra>0)
      createRelation(rseg,ArtePointer<MTRA>(iMtra),&RSEG::mtra);

    // -- seed based search: book RTRA-RSEG
    if (RingSearchTrack::exist && 
	RingSearchAttrib::seed_type==RingSearchType::rtra_seed) {
      createRelation(rseg, RingSearchTrack::rtra, &RSEG::rtra);
    }

    // -- book RSEG-HITB
    if (verb) cout<<"hitb="<<ArteTable<HITB>::size()
		  <<"  numhits="<<h.numhit
		  <<"  ring-hits="<<ring.numhit<<" nrings="<<numrings()<<endl;    
    if (RingSearchAttrib::bookhitb) {
      for (int i=0; i<ring.numhit; i++) {
	Hit &h=(*ring.hit[i]);
	if (!ring.assigned[i]) continue;
	createRelation(rseg, h.photon_hitb, &RSEG::hitb);
      }
    }


  }
}


void RingSearchARTE::FillRTRA(ArtePointer<RTRA> t, ArtePointer<RSEG> rich) {

  // -- fill likelihoods
  for (int i=0; i<n_mass; i++) {
    switch (i) {
    case 0: t->lre =lr[0];
    case 1: t->lrmu=lr[1];
    case 2: t->lrpi=lr[2];
    case 3: t->lrk =lr[3];
    case 4: t->lrp =lr[4];
    }
  }

  // -- fill number of rich hits
  int hri=0;
  for (ArteRelationIterator<HITB> ihitb=rich->hitb.begin();
       ihitb!=rich->hitb.end(); ++ihitb) hri++;
  t->hri=hri;

  //cout<<t<<":  lr="<<t->lre<<" "<<t->lrmu<<" "<<t->lrpi<<" "<<t->lrk<<" "<<t->lrp<<"   hri="<<t->hri<<endl;

}


/////////////////////// likelihood functions ///////////////////////////

void RingSearchARTE::makeR2P2lines() {
  int verb=0;
  if (verb) cout<<"RingSearchARTE::makeR2P2lines"<<endl;

  float  ipip = RingSearchTrack::ip*RingSearchTrack::ip;
  float  pp   = 1/ipip;
  static float mm[]  = {2.6e-7, 1.12e-2, 1.95e-2, .24372, .88035};
  float  sigma2_ipip = 4 * ipip*ipip * (RingSearchTrack::aa+RingSearchTrack::bb*pp);

  float r0   = ArtePointer<RITH>(1)->rad0;
  float r0r0 = r0*r0;
  for (int i=0; i<n_mass; i++) {
    iR2P2sigma2[i] = 1/(RingSearchTrack::sigma2_rr + mm[i]*mm[i] * sigma2_ipip);
    R2P2rad2[i]    = r0r0-mm[i]*ipip; 
    if (verb) cout<<"R2P2 lines "<<i<<": p="<<1/fabs(RingSearchTrack::ip)
		  <<"  r2="<<R2P2rad2[i]<<" sigma2="<<1/iR2P2sigma2[i]
		  <<" dp/p="<<sqrt(RingSearchTrack::aa+RingSearchTrack::bb*pp)
		  <<" sigma_ipip="<<sigma2_ipip<<" sigma2_rr="<<RingSearchTrack::sigma2_rr
		  <<" aa="<<RingSearchTrack::aa<<" bb="<<RingSearchTrack::bb<<endl;
  }
}



extern "C" float erfc_(float &x);

int RingSearchARTE::R2P2likelihood(ArtePointer<RTRA> t) {

  if (!RingSearchAttrib::id) return -1;

  // -- reset RTRA
  if (RingSearchAttrib::fillrtra) {
      if (RingSearchAttrib::use_simple_R2P2) 
	t->lre=t->lrmu=t->lrpi=t->lrk=t->lrp=10e4;
      else
	t->lre=t->lrmu=t->lrpi=t->lrk=t->lrp=0;
  }

  // -- find rich segment
  ArtePointer<RSEG> rich;
  ArteRelationIterator<RSEG> irseg=t->rseg.begin();
  for (;irseg!=t->rseg.end(); ++irseg) {
    if ((*irseg)->cmp==Rsegc::rich) { rich=(*irseg); break; }
  }
  if (irseg==t->rseg.end()) return -1;


  // -- fill number of rich hits
  int hri=0;
  if (RingSearchAttrib::fillrtra) { 
    for (ArteRelationIterator<HITB> ihitb=rich->hitb.begin();
	 ihitb!=rich->hitb.end(); ++ihitb) hri++;
    t->hri=hri;
  }
  
  // -- calculate likelihoods
  int ret=R2P2likelihood(rich->time);

  // -- fill tables
  if (RingSearchAttrib::fillrtra) FillRTRA(t,rich);

  return ret;
}



int RingSearchARTE::R2P2likelihood(float radius) {

  int verb=0;

  // -- calculate deviations from r2p2 lines
  static float chimax=50;
  float chi2=chimax;
  memset(&lr[0],0,sizeof(float)*n_mass);
  float rr=radius*radius;
  int bestM=-1;
  for (int i=0; i<n_mass; i++) {
    float d       = (rr - R2P2rad2[i]);
    lr[i]         = d*d*iR2P2sigma2[i];
    if (lr[i]<chi2) { chi2=lr[i]; bestM=i; }
  }

  if (verb) cout<<"r="<<radius<<" p="<<1/RingSearchTrack::ip<<endl;

  if (RingSearchAttrib::use_simple_R2P2) return bestM;

  static float isqrt2=1/sqrt(2);
  lrsum=0;
  if (bestM>-1) {
    // -- calculate probabilities
    for (int i=0; i<n_mass; i++) {
      if (verb) cout<<"likelihoods "<<i<<":  r2p2="<<lr[i];
      if (lr[i]>chimax) lr[i]=0;
      else {
	lr[i]=sqrt(lr[i])*isqrt2;
	lr[i]=erfc_(lr[i]);
	lrsum+=lr[i];
      }
      if (verb) cout<<"  lr="<<lr[i]
		    <<" (rr-r0r0)^2="<<(rr - R2P2rad2[i])*(rr - R2P2rad2[i])
		    <<"  R2P2r2="<<R2P2rad2[i]<<endl;
    }
    if (verb) cout<<endl;

    // -- renormalize
    if (lrsum>1e-6 && RingSearchAttrib::renorm) for (int i=0; i<n_mass; i++) lr[i]/=lrsum;
  }
  else memset(&lr[0],0,sizeof(float)*n_mass);

  return bestM;
}



RingSearchARTE *RICH_Search=0;




//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////// FUNCTIONS CALLED BY USER ////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

extern "C" {

void rcrings_(float r1,float r2,float dr,int t) {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings(r1,r2,dr,t);
}

void rcringsdb_() {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  RingSearchAttrib::bookhitb = true;
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings();
}

void rcringsid_(ArtePointer<RTRA> track) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  if (fabs(track->pf)>1) return;
  cout<<"rcringsid_"<<endl;
  RingSearchTrack::exist     = true;
  RingSearchTrack::rtra      = track;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  RingSearchAttrib::bookhitb = true;
  RICH_Search->getrings();
  RICH_Search->R2P2likelihood(track);
}

void rcringsclean_() {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = false;
  RingSearchAttrib::bookrseg = true;
  RingSearchAttrib::bookhitb = true;
  if (!RICH_Search) {
    fprintf(stderr," $$$ RICH cleanup: Ring cleanup called w/o previous call to ringseach\n");
    return;
  }
  RICH_Search->getrings();
}

	   }


float* ringid(ArtePointer<RTRA> track) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RingSearchTrack::ip        = track->pf;
  RingSearchAttrib::fillrtra = true;
  RICH_Search->makeR2P2lines();
  RICH_Search->R2P2likelihood(track);
  return RICH_Search->GetLr();
}

float* ringid(float pf, float radius) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RingSearchTrack::ip        = pf;
  RingSearchAttrib::fillrtra = false;
  RICH_Search->makeR2P2lines();
  RICH_Search->R2P2likelihood(radius);
  return RICH_Search->GetLr();
}

void ringsearchRTRA() {
  for(ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
      irtra!=ArteTable<RTRA>::end(); irtra++) {
    // -- do general ringsearch
    RingSearchTrack::max_phi_residual=.02;
    RingSearchTrack::max_lambda_residual=.075;
    //////RingSearchAttrib::fillrtra=true;
    RingSearchAttrib::type=RingSearchType::nobg;
    rcringsid_((*irtra));
  }  
  rcringsclean_();
}


void ringsearchRSEG() {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RingSearchAttrib::use_underdogs=false; /////////
  RingSearchTrack::max_phi_residual=.003;//////////
  RingSearchTrack::max_lambda_residual=.005;/////////////
  for(ArteTable<RSEG>::iterator irseg=ArteTable<RSEG>::begin();
      irseg!=ArteTable<RSEG>::end(); irseg++) {
    // -- choose pure PC
    if ((*irseg)->fit!=0) continue; 
    if (((*irseg)->cmp&(Rsegc::wch|Rsegc::patt))!=(Rsegc::wch|Rsegc::patt)) continue;
    //cout<<"rseg="<<(*irseg)<<" "<<hex<<(*irseg)->cmp<<dec<<endl;

    // -- do general ringsearch
    RingSearchTrack::exist      = true;
    RingSearchAttrib::seed_type = RingSearchType::rseg_seed;
    RingSearchTrack::rseg       = (*irseg);
    RingSearchAttrib::reuse     = false;
    RingSearchAttrib::bookrseg  = true;
    RingSearchAttrib::bookhitb  = true;
    RingSearchAttrib::id        = true;

    RingSearchAttrib::choose_best_ring  = false;
    RingSearchAttrib::finer_search_step = false;
    RingSearchAttrib::SNcut             = 4;

    RICH_Search->getrings();

    RingSearchAttrib::SNcut             = 3;

  }
  rcringsclean_();
}


void removerings() {
  for (ArteTable<RSEG>::iterator irseg=ArteTable<RSEG>::begin();
       irseg!=ArteTable<RSEG>::end(); ++irseg) {
    if ((*irseg)->cmp==Rsegc::rich) 
      (*irseg)->cmp=(Rsegc::bitsign|Rsegc::rich);
  }
}
