// ----------------------------------------------------------------------
//
// Fills MIMPs from the MIMP table into the HITB bank
// without using the DIGB bank.
//
// For the polarity, gede->xfw[2] is used.
//
// z-Positions for hit generation are the detector centers
// (which is correct for fully depleted detectors only!)
//
// Three parameters are used for hit generation:
//
// o  efficiency: Taken from 'silarte efficiency' IOL parameter
//                If < 0 (default) the GEDE efficency gede->eff is taken.
//
// o  smearing:   Taken from 'silarte hit resolution' IOL parameter
//                Default is 0.0030 = 30 um.
//
// o  double hit resolution:
//                Taken from 'silarte double hit resolution' IOL parameter
//                If < 0 (default), gede->dod is taken.
//
// o  noise:      Taken from 'silarte noise rate' IOL parameter.
//                Default is 0.
//                The number of noise hits per detector side is
//                poisson distributed around the mean 
//                        noise-rate * #(strips)
//
// Note, that efficiency and double hit resolution are only detector
// dependent, if taken from the GEDE tables! So, even if you want to
// change those values, its probably better to update the GEDE tables
// than to use an overall value by using the respective IOL parameter.
//
// Author: Rainer Wanke, 25-Sep-1997
//         Modifications by Silvia Masciocchi, Dec. 1998
//         Modifications for new C++ interface by RW, 3 Feb 99
//         Introduction of noise by RW, 18 May 99
//
// ----------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <set>
#include <multiset.h>

// Arte banks,  C++ version
#include "arte/arte.hh"
#include "silarte/silarte.hh"
#include "silarte/EVinfo.hh"

#include "aframelib.h"
#include "grr.h"

// HITB-MIMP pair structure
typedef pair< HITB, ArtePointer<MIMP> > MimpHit;

// strip counts of n and p sides
#define  NSTRIPN   1280
#define  NSTRIPP   1024

// cernlib gaussian/poisson random number generators and 3x3 matrix inversion
extern "C" {
  extern void rnorml_( float *, int * );
  extern void rnpssn_( float *, int *, int * );
  extern void rinv_( int *, float *, int *, int *, int * );
}
#define rnorml(A)   do { float* a=(A);          \
                         int   one=1;           \
                         rnorml_( a, &one ); }  \
                    while (1==0)
#define rnpssn(A,B) do { float  a=(A);                \
                         int    err;                  \
                         rnpssn_( &a, (B), &err ); }  \
                    while (1==0)
#define r3inv(A,B) do { int array[3];                                \
                        int three=3;                                 \
                        rinv_( &three, (A), &three, array, (B) ); }  \
                   while (1==0)

// ----------------------------------------------------------

// function returns true for hit1 < hit2 
struct LessMimpHit: public binary_function<MimpHit,MimpHit,bool>
{
  bool operator()( const MimpHit & hit1, const MimpHit & hit2 )
    {
      if ( hit1.first.gede != hit2.first.gede )
	return hit1.first.gede < hit2.first.gede;
      
      return hit1.first.x < hit2.first.x;
    }
};


// function prototypes
void createHITBfromMIMP(  ArtePointer<GEDE> gede, 
			  ArtePointer<MIMP> mimp, 
			  multiset< MimpHit, LessMimpHit > *hits );
void createHITBfromNoise( multiset< MimpHit, LessMimpHit > *hits );
int silarte_updateGEDE_MC();

// ----------------------------------------------------------

int fillHITBfromMIMP()
{
  ArtePointer<GEDE> oldgede = 0;

  static int first = 1;

  // multiset for storing hits and their MIMPs 
  multiset< MimpHit, LessMimpHit > hits;

  char mess[256];
  char *routine = "fillHITBfromMIMP";

  // --------------------------------------------------------

  // Check, if we have a MC event
  if ( !CheckMC() )
    return GRR_OK;

  // modify zfw[0] in GEDE tables, if first event
  if ( first ) {
    silarte_updateGEDE_MC();
    first = 0;
  }

  // loop over MIMPs
  for ( ArteTable<MIMP>::iterator mimpIt = ArteTable<MIMP>::begin();
	mimpIt != ArteTable<MIMP>::end(); mimpIt++ ) {
    ArtePointer<MIMP> mimp = *mimpIt;

    // GEDE bank for this MIMP entry (this is the SVD front side)
    ArtePointer<GEDE> gedefront = mimp->gede;
    ArtePointer<GEDE> gedeback  = 0;

    // first check, if MIMP is a silicon MIMP at all
    if ( gedefront->cmp == 1 ) { 

      // find GEDE bank for the detector's back side
      for ( ArteTable<GEDE>::iterator gedeIt = ArteTable<GEDE>::begin();
	    gedeIt != ArteTable<GEDE>::end(); gedeIt++ ) {
	ArtePointer<GEDE> gede = *gedeIt;
        if (    gede->cmp  == 1
	     && gede->sec  == gedefront->sec
	     && gede->gesl == gedefront->gesl ) {
	  if (   gede->lay == 1 && gedefront->lay == 2
              || gede->lay == 2 && gedefront->lay == 1
              || gede->lay == 3 && gedefront->lay == 4
	      || gede->lay == 4 && gedefront->lay == 3 )
	    gedeback = gede;
	}
      }

      // print out this first success if in debug mode
      if ( silarte_iol.debug_mode.value )
	aframe_mess(routine,GRR_SINFO,
		    "MIMP %d at z = %f, x = %f, y = %f:",
		    mimp.index(), mimp->z, mimp->x, mimp->y );

      // create front side hits
      createHITBfromMIMP( gedefront, mimp, &hits );
	
      // create back side hits
      createHITBfromMIMP( gedeback, mimp, &hits );
	
    }
    
  }   // end loop over MIMPs

  // create noise hits if requested
  createHITBfromNoise( &hits );

  // Now fill HITB table with the accumulated hits
  ArtePointer<HITB> last_hitb;
  int   ndouble = 0; 
  float xsum    = 0.;
  for ( multiset< MimpHit, LessMimpHit >::iterator ihit = hits.begin();
        ihit != hits.end(); ihit++ ) {
    
    // first check for double hits - they are merged
    if ( ihit != hits.begin() ) {
      float dhres = silarte_iol.double_hit_resolution.value;
      if ( dhres < 0 )
	dhres = ihit->first.gede->dod;

      // if x is to close to last hit, just add another MIMP relation
      if (   ihit->first.gede == last_hitb->gede
	  && ihit->first.x - last_hitb->x < dhres ) {

	// x-values for double hits are added up...
	if ( ndouble == 0 ) {
	  xsum += last_hitb->x;
	  ndouble++;
	  if ( silarte_iol.debug_mode.value )
	    sprintf(mess,"Double hits in GEDE %d: %f",
		    last_hitb->gede.index(), last_hitb->x );
	}
	xsum += ihit->first.x;
	ndouble++;
	if ( silarte_iol.debug_mode.value )
	   sprintf( mess, "%s, %f", mess, ihit->first.x );

	// create another HITB->MIMP relation for double hit
	ArtePointer<MIMP> mimp = ihit->second;
	if ( mimp != 0 )
	  createRelation( last_hitb, mimp, &MIMP::hitb );

	continue;
      }

      // Modify x-value if last hit turned out to be a double hit
      else if ( ndouble > 0 ) {
	last_hitb->x = xsum / (float) ndouble; 
	ndouble = 0;
	xsum    = 0.;
	if ( silarte_iol.debug_mode.value )
	   aframe_mess(routine,GRR_SINFO,
		       "%s ==> %f", mess, last_hitb->x );
      }
    }

    // fill hit and its mimp relation into HITB table
    ArtePointer<HITB> hitb = ArteTable<HITB>::push_back( ihit->first );
    ArtePointer<MIMP> mimp = ihit->second;
    if ( mimp != 0 )
      createRelation( hitb, mimp, &MIMP::hitb );
    
    // Fill GEDE hit1 and hitl
    if ( hitb->gede != oldgede ) {
      hitb->gede->hit1 = hitb;
      oldgede = hitb->gede;
    }
    hitb->gede->hitl = hitb;

    last_hitb = hitb;
  }

  // Modify x-value of last hit if it was a double hit
  if ( ndouble > 0 ) {
    last_hitb->x = xsum / (float) ndouble; 
    if ( silarte_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,
		  "%s ==> %f", mess, last_hitb->x );
  }

  return GRR_OK;
  
}

// ----------------------------------------------------------
//
// fill MIMP into MimpHit multiset-structure
//
void createHITBfromMIMP( ArtePointer<GEDE> gede, 
			 ArtePointer<MIMP> mimp, 
			 multiset< MimpHit,LessMimpHit > *hits )
{
  char *routine = "createHITBfromMIMP";

  // check for existance of GEDE (could be single-sided detector)
  if ( gede == 0 ) {
    if ( silarte_iol.debug_mode.value )
      aframe_mess(routine,GRR_SERROR,
		  "   One GEDE side missing for MIMP %d", mimp.index() );
    return;
  }

  if ( silarte_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"   GEDE %d: sl = %s, sec = %d, lay = %d",
		gede.index(), gede->gesl->name, gede->sec, gede->lay );

  // correct for finite thickness of detector
  // hits are generated with respect to the center
  // of the detector (correct for fully depleted detectors!)
  float delta_z = (gede->zmax + gede->zmin) / 2. - mimp->z;
  float x       = mimp->x + mimp->tx * delta_z;
  float y       = mimp->y + mimp->ty * delta_z;
  float z       = (gede->zmax + gede->zmin) / 2.;

  // check that the MIMP hits the sensitive area
  // First rotate the MIMP back (easier than rotating the detector)
  // 1. Invert the rotation matrix
  float rinv[9];
  for ( int i=0; i<9; i++ )
    rinv[i] = gede->rot[i];
  int ifail = 0;
  r3inv( rinv, &ifail );   // invert rotation matrix
  if ( ifail == -1 ) {
    aframe_mess(routine,GRR_SERROR,
		"Rotation matrix not invertable for GEDE %d",
		gede.index() );
    return;
  }
  // 2. Mulitiply MIMP with inverted rotation matrix
  float xinv = x * rinv[0] + y * rinv[1] + z * rinv[2];
  float yinv = x * rinv[3] + y * rinv[4] + z * rinv[5];
  if ( silarte_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"     Rotated (x,y) = (%f,%f) to (x',y') = (%f,%f)",
		x, y, xinv, yinv );

  // At the moment we don't care about the insensitive corner on the n-side 
  if (   xinv < gede->sxmn || xinv > gede->sxmx 
      || yinv < gede->symn || yinv > gede->symx ) {
    if ( silarte_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,
		  "     MIMP for GEDE %d outside sensitive area!",
		  mimp->gede.index() );
    return;
  }

  // take into account the detector efficiency
  float pf  = (float) rand() / (float) RAND_MAX;
  float eff = silarte_iol.efficiency.value;
  if ( eff < 0 )
    eff = gede->eff;
  if ( pf > eff ) {
    if ( silarte_iol.debug_mode.value )
       aframe_mess(routine,GRR_SINFO,
		   "      Lost MIMP for GEDE %d because of inefficiency.",
		   gede.index() );
    return;
  }

  // Check which side it is, to use the right sign for the charge
  float signb = gede->xfw[2];
	  
  // Convert x,y,z into HITB clusters
  // z-coordinate is center of detector
  HITB hit;
  hit.z    = (gede->zmax + gede->zmin) / 2.;
  hit.x    = x * gede->cos - y * gede->sin;
  hit.y    = 0.;
  hit.q    = 25. * signb;
  hit.r    = silarte_iol.hit_resolution.value;  
  hit.gede = gede;
  
  // Do smearing
  float smear;
  rnorml( &smear );
  hit.x = hit.x + smear * silarte_iol.hit_resolution.value;
  
  if ( silarte_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"      HITB: x = %f (smeared by %f um), z = %f, GEDE = %d",
		hit.x, smear * silarte_iol.hit_resolution.value * 10000.,
		hit.z, hit.gede.index() );
  
  // Fill hit into multiset
  hits->insert( make_pair( hit, mimp ) );

  return;
}

// ----------------------------------------------------------
//
// fill noise hits into MimpHit multiset-structure
//
void createHITBfromNoise( multiset< MimpHit,LessMimpHit > *hits )
{
  char *routine = "fillHITBfromNoise";

  // check correct noise value ( 0 < noise rate <= 1 ) 
  if ( silarte_iol.noise_rate.value <= 0. )
    return;
  if ( silarte_iol.noise_rate.value > 1. )
    silarte_iol.noise_rate.value = 1;

  // loop over GEDE tables = detector sides
  // --------------------------------------
  for ( ArteTable<GEDE>::iterator gedeIt = ArteTable<GEDE>::begin();
	gedeIt != ArteTable<GEDE>::end(); gedeIt++ ) {
    ArtePointer<GEDE> gede = *gedeIt;
    if ( gede->cmp != 1 )
      continue;

    // determine number of strips of side from pitch
    // (Note: Single sided detectors are p-sides with n-side geometry!)
    int nstrip;
    if ( gede->pit < 0.0053 )    // n side type
      nstrip = NSTRIPN;
    else                         // p side type
      nstrip = NSTRIPP;

    // find poisson-distributed number of noise clusters
    int nnoisehit;
    rnpssn( nstrip * silarte_iol.noise_rate.value, &nnoisehit );
    if ( silarte_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO, "Creating %d noise hits.", nnoisehit );

    // Check which side it is, to use the right sign for the charge
    float signb = gede->xfw[2];

    // loop over noise hit count to create the noise
    for ( int i=0; i<nnoisehit; i++ ) {

      // randomly distributed strip-number
      float cog = (float) rand() / (float) RAND_MAX * (float) nstrip;
 
      //  HITB clusters
      // z-coordinate is center of detector
      HITB hit;
      hit.z    = (gede->zmax + gede->zmin) / 2.;
      hit.x    = cog * gede->pit + gede->xfw[0];
      hit.y    = 0.;
      hit.q    = 10. * signb;
      hit.r    = silarte_iol.hit_resolution.value;  
      hit.gede = gede;
  
      if ( silarte_iol.debug_mode.value )
	aframe_mess(routine,GRR_SINFO,
		    "      Noise HITB: x = %f, z = %f, GEDE = %d",
		    hit.x, hit.z, hit.gede.index() );
      
      // Fill hit without a mimp into multiset
      hits->insert( make_pair( hit, 0 ) );
      
    }

  }  // end loop over GEDE table

  return;
}

// ----------------------------------------------------------
//
// set GEDE->zfw[0] to center of detector = (zmin+zmax) / 2
//
int silarte_updateGEDE_MC()
{
  char *routine = "silarte_updateGEDE_MC";

  for ( ArteTable<GEDE>::iterator igede = ArteTable<GEDE>::begin();
	igede != ArteTable<GEDE>::end(); igede++ ) {
    ArtePointer<GEDE> gede = *igede;

    if ( gede->cmp == 1 )
      gede->zfw[0] = ( gede->zmax + gede->zmin ) / 2.;

  }

  if ( silarte_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"GEDE table updated for correct SVD-hit z-positions.");

  return GRR_OK;
}
