// *****************************************************************************
//
// source:    L3_Chrono.C
//
// type:      source code
//
// created:   06 Dec 1999
//
// author:    Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// symbol:    
//
// changes:
//
// *****************************************************************************
#undef TESTCHRONO

#include "L3_Chrono.hh"
#include <unistd.h>
#include <sys/times.h>
#include <iostream.h>
#include <string.h>

// Static member
//--  Conversion factor from clock ticks to seconds
//    time(seconds) = rclk_tck * time(clock_ticks)
float L3_Chrono::rclk_tck = 0.0; 

// Constructors: they need a name for the Chrono that can be defined
// both by a string or by an array of characters
  
L3_Chrono::L3_Chrono(char *s) {
  if( strlen(s)<14){
    strcpy(chronoName,s);
  } else {
    strncpy(chronoName,s,13);
    chronoName[13] =0;
  }
  if( rclk_tck==0.0 )   // get the relation clock ticks- seconds
    rclk_tck = 1./sysconf(_SC_CLK_TCK); 
  reset();
}
/*
L3_Chrono::L3_Chrono(string &s) : chronoName(s) {
  if( rclk_tck==0.0 )   // get the relation clock ticks- seconds
    rclk_tck = 1./sysconf(_SC_CLK_TCK); 
  reset();
}
*/
//Destructor: now is empty
L3_Chrono::~L3_Chrono(){}

// Reset operations
void L3_Chrono::reset(){
  numberOfMeasurements = 0;
  total_utime = 0;
  total_stime = 0;
  start_utime = (clock_t) -1;
  start_stime = 0;
}

// Start:  Start a measuring time operation
void L3_Chrono::start(){
  struct tms measuredTime;
  if(start_utime!= (clock_t)-1)
    cerr<<" L3_Chrono "<<chronoName<<
      " error: received a Start message without a preceding Stop one!"<<endl;
  // get the system times for the current process
  times(&measuredTime);
  // Store in object the measured times
  start_utime = measuredTime.tms_utime;
  start_stime = measuredTime.tms_stime;
}


// Stop:  Stop a measuring time operation and update internal counters
void L3_Chrono::stop(){
  struct tms measuredTime;
  // test start_utime to see if there is a measurement in progress  
  if(start_utime==(clock_t) -1){
    // no measurement in progress!!!
    cerr<<" L3_Chrono "<<chronoName<<
      " error: received a Stop message without a preceding Start one!"<<endl;
  } else {
    // get the system times for the current process
    times(&measuredTime);
    // Evaluate elapsed times
    clock_t elapsed_time;
    // evaluate for system time
    elapsed_time = measuredTime.tms_stime - start_stime;
    total_stime += elapsed_time;
    start_stime = elapsed_time;
    // evaluate for user time
    elapsed_time = measuredTime.tms_utime - start_utime;
    total_utime += elapsed_time;
    start_utime = (clock_t) -1;
    start_stime += elapsed_time;
    // update number of measurements made
    numberOfMeasurements++;
    // after this point in total_xtime there are values relative to
    // numberOfMeasurements measures; in start_utime there is -1 (that is
    // a flag for waiting a Start message) and in start_stime there is
    // the last measured time (user+system)
  }
}


//   Print all information for the current Chrono
void L3_Chrono::print() const {
  float  totaltime,usertime,systemtime,avtottime,avustime,avsystime;
  evaluateSeconds(totaltime,usertime,systemtime,
		  avtottime,avustime,avsystime);

  // print out
  cout<<"L3_Chrono: "<<chronoName<<"     \tMeasurements :"
      <<numberOfMeasurements<<endl;
  cout<<"Total   time "<<totaltime<<" s  \tUser: "<<usertime
      <<" s \tSystem "<<systemtime<<" s"<<endl;
  if(numberOfMeasurements>0){
    cout<<"Average time "<<avtottime<<" s  \tUser: "<<avustime
	<<" s \tSystem "<<avsystime<<" s"<<endl;
  }
}

//   Print all infos in a single line
void L3_Chrono::printLine() const {
  float  totaltime,usertime,systemtime,avtottime,avustime,avsystime;

  evaluateSeconds(totaltime,usertime,systemtime,
		  avtottime,avustime,avsystime);

  cout<<chronoName;
  for(int i=15-strlen(chronoName); i>0; i--)
    cout<<" ";
  cout.width(4);

  cout.setf(ios::fixed);
  //  cout.precision(4);
  cout<<numberOfMeasurements<<" \t"
      <<totaltime<<"  "<<usertime<<"  "<<systemtime<<"  "
      <<avtottime<<"  "<<avustime<<"  "<<avsystime<<endl;
  /*  cout.form("%5d %7.3f %7.3f %7.3f   %7.4f %7.4f %7.4f\n",
 	   numberOfMeasurements,
           totaltime,usertime,systemtime,
           avtottime,avustime,avsystime);
	   */
}

//    Evaluate times in seconds: private member function
void  L3_Chrono::evaluateSeconds(float & totaltime, float & usertime,
			      float & systemtime,float & avtottime,
			      float & avustime, float &avsystime) const {
  // evaluate times in seconds
  totaltime = (total_utime+total_stime)*rclk_tck;
  usertime  = total_utime * rclk_tck;
  systemtime= total_stime * rclk_tck;
  if(numberOfMeasurements>0){
    // evaluate average times
    float r = 1./numberOfMeasurements;
    avtottime  = totaltime*r;
    avustime   = usertime*r;
    avsystime = systemtime*r;
  } else {
    avtottime = 0.0;
    avustime  = 0.0;
    avsystime = 0.0;
  }
}



#ifdef TESTCHRONO
main(){
  L3_Chrono  outer("Outer loop");
  L3_Chrono  inner("Inner loop");
  L3_Chrono  medium("Medium loop");
  
  int i,j;
  outer.start();
  for(i=0; i<200; i++){
     medium.start();
     for(j=0; j<100; j++){
       inner.start();
       inner.stop();
     }
     medium.stop();
  }
  outer.stop();

  outer.print();
  medium.print();
  inner.print();  

}
#endif

